use std::{fs::File, io::Read, path::{Path, PathBuf}, process::Command};

use viperid::VResult;

pub struct GoBuilder {
    project_directory: PathBuf,
    entry_point: PathBuf
}

impl GoBuilder {
    pub fn new(project_directory: &Path) -> VResult<GoBuilder> {
        // make sure Go is a supported version
        let output = Command::new("go").arg("version").output()?;

        if !output.status.success() {
            anyhow::bail!("Go is not installed or did not start: {:?}", output.status)
        }

        // TODO: Support more Go versions
        let text = output.stdout;
        if !text.starts_with(b"go version go1.21") {
            anyhow::bail!("unsupported Go version (must be 1.21): {}", String::from_utf8_lossy(&text).trim())
        }

        // make sure the project is really here
        if !project_directory.exists() {
            anyhow::bail!("project directory must exist: {}", project_directory.to_string_lossy())
        }

        let entry_point = project_directory.join("main.go");
        if !entry_point.exists() {
            anyhow::bail!("entry point must exist: {}", entry_point.to_string_lossy())
        }

        // OK, we should be able to do things with Go without any errors
        Ok(GoBuilder {
            project_directory: PathBuf::from(project_directory),
            entry_point
        })
    }

    pub fn build(&self) -> VResult<Vec<u8>> {
        // NOTE: We've checked that the project directory exists
        if !self.project_directory.exists() {
            anyhow::bail!(
                "project directory disappeared: {}", 
                self.project_directory.to_string_lossy()
            );
        }
        if !self.entry_point.exists() {
            anyhow::bail!(
                "entry point disappeared: {}", 
                self.entry_point.to_string_lossy()
            );
        }

        let project_directory = &self.project_directory;
        let entry_point = &self.entry_point;
        let build_directory = project_directory.join("build");
        let wasm_name = build_directory.join("game.wasm");

        let mut child = Command::new("go")
            .args(["build", "-o"]).arg(&wasm_name)
            .env("GOOS", "wasip1")
            .env("GOARCH", "wasm")
            .arg("-trimpath")
            .arg(entry_point)
            .spawn()?;
        let status = child.wait()?;
        if !status.success() {
            anyhow::bail!("go failed");
        }

        let mut buf = vec![];
        File::open(wasm_name)?.read_to_end(&mut buf)?;
        Ok(buf)
    }


}