use std::{fmt::Display, mem};

use viperid::VResult;
use wasmi::{core::{HostError, Trap}, Engine, Linker, Module, Store, TypedFunc, TypedResumableCall, TypedResumableInvocation};

use crate::{engine_api::{self, YieldFrame}, wasi::{self, StockWasi}};

pub struct Executor {
    store: Store<ExecutorState>,
    resumable: Resumable
}

pub(crate) struct ExecutorState {
    wasi: wasi::StockWasi
}

impl ExecutorState {
    fn new() -> Self {
        ExecutorState {
            wasi: StockWasi::new()
        }

    }
}


impl Executor { 
    pub fn new(wasm: &[u8]) -> VResult<Self> {
        let engine = Engine::default();

        let module = Module::new(&engine, wasm)?;

        let mut store = Store::new(&engine, ExecutorState::new());

        let mut linker = <Linker<ExecutorState>>::new(&engine);
        wasi::integrate(&mut linker, |hs| &mut hs.wasi)?;
        engine_api::integrate(&mut linker);

        let instance = linker
            .instantiate(&mut store, &module)?
            .start(&mut store)?;

        let main = instance.get_typed_func::<(), ()>(&store, "_start")?;

        Ok(Executor {
            store,
            resumable: Resumable::Unstarted(main)
        })
    }

    pub fn is_running(&self) -> bool {
        self.resumable.is_running()
    }

    pub fn get_error(&self) -> VResult<()> {
        if let Err(e) = self.resumable.get_error() {
            anyhow::bail!("executor error: {}", e);
        }
        Ok(())
    }

    pub fn update(&mut self) {
        let mut tmp = Resumable::Failed(wasmi::Error::Trap(Trap::new("unexpected crash during update")));
        mem::swap(&mut tmp, &mut self.resumable);
        let mut tmp2 = tmp.resume(&mut self.store);
        mem::swap(&mut tmp2, &mut self.resumable);
    }
}

enum Resumable {
    Unstarted(TypedFunc<(), ()>),
    Paused(TypedResumableInvocation<()>),
    Failed(wasmi::Error),
    Completed(())
}


impl Resumable {
    pub fn resume(self, store: &mut Store<ExecutorState>) -> Resumable {
        let result: Result<TypedResumableCall<()>, wasmi::Error> = match self {
            Resumable::Failed(f) => return Resumable::Failed(f),
            Resumable::Completed(c) => return Resumable::Completed(c),
            Resumable::Unstarted(u) => {
                match u.call_resumable(store, ()) {
                    Ok(o) => Ok(o),
                    Err(e) => Err(wasmi::Error::Trap(e))
                }
            }
            Resumable::Paused(p) => {
                let trap = p.host_error();

                if let Some(_) = trap.downcast_ref::<YieldFrame>() {
                    p.resume(store, &[])
                } else {
                    Err(wasmi::Error::Trap(Trap::new(format!("unexpected host error: {}", trap))))
                }
            }
        };

        match result {
            Ok(o) => match o {
                TypedResumableCall::Finished(f) => Resumable::Completed(f),
                TypedResumableCall::Resumable(p) => Resumable::Paused(p)
            }
            Err(e) => Resumable::Failed(e),
        }
    }
    
    fn is_running(&self) -> bool {
        match self {
            Resumable::Unstarted(_) => true,
            Resumable::Paused(_) => true,
            Resumable::Failed(_) => false,
            Resumable::Completed(_) => false
        }
    }
    
    fn get_error(&self) -> Result<(), &wasmi::Error> {
        match self {
            Resumable::Unstarted(_) => (),
            Resumable::Paused(_) => (),
            Resumable::Failed(e) => return Err(e),
            Resumable::Completed(_) => ()
        };
        Ok(())
    }
}
